/*********************************************************************
	created:	2003/04/27
	created:	27:4:2003   17:51
	filename: 	E:\Microsoft Visual Studio\MyProjects\extgen\extgen.cpp
	file path:	E:\Microsoft Visual Studio\MyProjects\extgen
	file base:	extgen
	file ext:	cpp
	author:		Andreas Kirsch
	
	purpose:	This programm will offer a general algorithm to use
				the OpenGL extension with Win32.
				It will use the STL

***********************************************************************/
#include "stdio.h"
#include <vector>
#include <string>

//////////////////////////////////////////////////////////////////////////
// parser functions and declarations
//////////////////////////////////////////////////////////////////////////
using namespace std;

struct Extension 
{
	string name; // extension name

	vector<string> function_list; // these are the function prototypes
};

struct Registry
{
	string extension_prefix;
	string function_prefix;
	string pointer_prefix;
	string prototype;

	int total_functions;
	vector<Extension> extension_list;

	FILE *input;
	FILE *header;
	
	void OpenHeader(char *file);
	void CreateOutputHeader(char *file);
	
	void Parse_Header();
	void Write_Header();
	void List();

	Registry(const char *fprefix) :function_prefix(fprefix), total_functions(0) {}
};

int GetToken(FILE *input, char *token)
{
	return fscanf(input,"%s",token);
}

int GetFunction(char *token, char *out)
{
	return sscanf(token,"%[^ (\t\n]",out);
}


string& ToUpper(string& str)
{
	for(size_t i=0; i < str.size(); i++)
		if(str[i]>= 'a' && str[i] <= 'z')
			str[i]+='A'-'a';

	return str;
}

string& ToLower(string& str)
{
	for(size_t i=0; i < str.size(); i++)
		if(str[i]>= 'A' && str[i] <= 'Z')
			str[i]-='A'-'a';
		
		return str;
}

// open the header
void Registry::OpenHeader(char *file)
{
	input=fopen(file,"r");
	if(input == NULL)
	{
		printf("Failed opening %s\n",file);
		exit(1);
	}

}

void Registry::CreateOutputHeader(char * file)
{
	header = fopen(file,"w");
	if(header == NULL)
	{
		printf("Failed creating %s\n",file);
		exit(1);
	}
}

void Registry::List()
{
	printf( "* Information :\n"
			"*\n"
			"* Extension type : %s\n"
			"*\n"
			"* Extensions found : %u  Functions  found : %u\n"
			"\n",
			function_prefix.c_str(),extension_list.size(),total_functions);

	printf(	"** Extension List **\n"
			"* Extension Name : Function Count\n\n");

	for(size_t ext_list = 0; ext_list < extension_list.size(); ext_list++)
		printf("%-30s : %3i\n",extension_list[ext_list].name.c_str(),
			   extension_list[ext_list].function_list.size());

	printf(	"\n"
			"** Structured View **\n"
			"\n");

	for(size_t ext = 0; ext < extension_list.size(); ext++)
	{
		printf(	"%s\n"
				"{\n",extension_list[ext].name.c_str());

		for(size_t func = 0; func < extension_list[ext].function_list.size(); func ++)
			printf("\t%s\n",extension_list[ext].function_list[func].c_str());

		printf( "}\n\n");
	}
}

// creates all extension entries
void Registry::Parse_Header()
{
	char token[256];
	char function_token[256];
	
	// build helper strings, etc.
	
	ToUpper(prototype = function_prefix);
	prototype = prototype + "_" + prototype+"EXT_PROTOTYPES";

	pointer_prefix = "PFN";

	ToUpper(extension_prefix = function_prefix)+="_";

	// temp strings;
	string tempstr;

	Extension ext;

	// start scanning
	GetToken(input,token);
	
	while(!feof(input))
	{

		// repeat until the first #ifndef
		if(strcmp(token,"#ifndef") != 0)
		{
			GetToken(input,token);
			continue;
		}
		
		GetToken(input,token);

		// has the macro a extension_prefix ?
		if(string(token,extension_prefix.size()) != extension_prefix)
			continue;

		// save it
		tempstr = token;

		// look for the #ifdef ( break if you find an #endif )
		while (1) 
		{
			GetToken(input,token);
			
			if(strcmp(token,"#ifdef") == 0 )
				break;
			if(strcmp(token,"#endif") == 0)
				break;
		}

		// has the next token a prototype macro  ?
		GetToken(input,token);
		if(token != prototype)
			continue;

		// create a new extension
		ext.name=tempstr;

		// enumerate every function
		while(1)
		{
			GetToken(input,token);
			if(	string(token,function_prefix.size()) == function_prefix)
			{
				// add new function entry
				
				// get the only the function name
				GetFunction(token,function_token);

				ext.function_list.push_back(string(function_token));

				total_functions++;
				//go on
				continue;
			}

			if(strcmp(token,"#endif") == 0)
				break;
		}

		extension_list.push_back(ext);

		// reset ext
		ext.function_list.clear();

		GetToken(input, token);
	}
}

// writes every function template into a new header file
void Registry::Write_Header()
{
	const char * fprefix = function_prefix.c_str();
	const char * pprefix =  prototype.c_str();
	const char * p_prefix = pointer_prefix.c_str();
	string upfprefix = function_prefix;

	ToUpper(upfprefix);

	string tempstr;

//////////////////////////////////////////////////////////////////////////
// create header entries
//////////////////////////////////////////////////////////////////////////

	fprintf(header,"// created with (W)GLextgen by Andreas Kirsch 2003\n\n");
	fprintf(header,"// once you must #define GL_EXT_INIT_SOURCE/WGL_EXT_INIT_SOURCE\n");
	fprintf(header,"// before including this header.\n\n");
	
	if(function_prefix == "gl")
		fprintf(header, "// To display a list of the gl versions supported by this header\n"
						"// in the debug window define GL_LIST_SUPPORTED_VERSIONS\n");
	
	fprintf(header,"// To display all supported extensions define %sLIST_ALL_EXT\n",
					extension_prefix.c_str());

	// instead of #pragma once, we use a common #ifdef
	fprintf(header,"#ifndef __FILE_%s_EXT_PROC\n", upfprefix.c_str());
	fprintf(header,"#define __FILE_%s_EXT_PROC\n\n", upfprefix.c_str());

	fprintf(header,"// include the default opengl header files\n");
	fprintf(header,"#include <GL\\gl.h>\n");
//	fprintf(header,"#include <GL\\glu.h>\n");
	// include wglext.h or glext.h
	fprintf(header,"#include <GL\\%sext.h>\n\n",function_prefix.c_str());

	fprintf(header,"#ifdef %s_EXT_INIT_SOURCE\n",upfprefix.c_str());

	// include windows.h because of wglGetProcAddress and string.h because of strcmp
	fprintf(header,"#\tinclude \"windows.h\"\n");
	fprintf(header,"#\tinclude \"string.h\"\n");
		
	fprintf(header,"#endif //%s_EXT_INIT_SOURCE\n",upfprefix.c_str());

	// extern "c" decl
	fprintf(header,"#ifdef __cplusplus\n");
	fprintf(header,"extern \"C\"\n");
	fprintf(header,"{\n");
	fprintf(header,"#endif\n\n");
	
	fprintf(header,"GLboolean _%sExtInit(const char *extension);\n\n",fprefix);

	for(size_t extension = 0; extension < extension_list.size(); extension++)
	{
		fprintf(header,"#ifdef %s\n",extension_list[extension].name.c_str());
		
		// GL_LIST_SUPPORTED_VERSIONS and (W)GL_LIST_ALL_EXT added [4/28/2003]
		if(function_prefix == "gl" && extension_list[extension].name.size() > 11
		   && string(extension_list[extension].name.c_str(),11) == "GL_VERSION_")
			fprintf(header,	"#if defined(GL_LIST_SUPPORTED_VERSIONS) || defined"
							"(%sLIST_ALL_EXT)\n"
							"#\tpragma message(\"%s supported\")\n"
							"#endif\n", extension_prefix.c_str(),
							extension_list[extension].name.c_str());
		else
			fprintf(header,	"#if defined(%sLIST_ALL_EXT)\n"
							"#\tpragma message(\"%s supported\")\n"
							"#endif\n", extension_prefix.c_str(),
							extension_list[extension].name.c_str());
		
		for(size_t function = 0; function < extension_list[extension].function_list.size();function++)
		{
			ToUpper(tempstr = extension_list[extension].function_list[function]);
			fprintf(header,"\t%s%sPROC %s = 0;\n",p_prefix,tempstr.c_str(),
				extension_list[extension].function_list[function].c_str());
		}
/*		for(int function = 0; function < extension_list[extension].function_list.size();function++)
		{
			ToUpper(tempstr = extension_list[extension].function_list[function]);
			fprintf(header,"\t%s%sPROC _EXT_%s_ = 0;\n",p_prefix,tempstr.c_str(),
					extension_list[extension].function_list[function].c_str());
		}
		fprintf(header,"\n");
		for(int dfunction = 0; dfunction < extension_list[extension].function_list.size()
			;dfunction++)
		{
			ToUpper(tempstr = extension_list[extension].function_list[dfunction]);
			fprintf(header,"\t#define %s _EXT_%s_\n",
				extension_list[extension].function_list[dfunction].c_str(),
				extension_list[extension].function_list[dfunction].c_str());
		}*/
		fprintf(header,"#endif //%s\n\n", extension_list[extension].name.c_str());
	}

//////////////////////////////////////////////////////////////////////////
// create _glExtInit()/_wglExtInit()
//////////////////////////////////////////////////////////////////////////
	
	fprintf(header,"#ifdef %s_EXT_INIT_SOURCE\n\n",upfprefix.c_str());

	fprintf(header,"GLboolean _%sExtInit(const char *extension)\n",fprefix);
	fprintf(header,"{\n");
	
	fprintf(header,"\tif(extension == 0)\n");
	fprintf(header,"\t\treturn GL_FALSE;\n");

	for(size_t ext = 0; ext < extension_list.size(); ext++)
	{
		fprintf(header,"#ifdef %s\n",extension_list[ext].name.c_str());

		fprintf(header,"\telse if(0 == strcmp(\"%s\",extension))\n",
		extension_list[ext].name.c_str());
		
		fprintf(header,"\t{\n");
		
		for(size_t func = 0; func < extension_list[ext].function_list.size(); func++)
		{
			ToUpper(tempstr = extension_list[ext].function_list[func]);
			
/*			fprintf(header,"\t\t_EXT_%s_ = (%s%sPROC)wglGetProcAddress(\"%s\");\n",
				extension_list[ext].function_list[func].c_str(),p_prefix,tempstr.c_str(),
				extension_list[ext].function_list[func].c_str());
			fprintf(header,"\t\tif(_EXT_%s_ == 0)\n",
					extension_list[ext].function_list[func].c_str());
			fprintf(header,"\t\t\treturn GL_FALSE;\n");*/
			fprintf(header,"\t\t%s = (%s%sPROC)wglGetProcAddress(\"%s\");\n",
				extension_list[ext].function_list[func].c_str(),p_prefix,tempstr.c_str(),
				extension_list[ext].function_list[func].c_str());
			fprintf(header,"\t\tif(%s == 0)\n",
				extension_list[ext].function_list[func].c_str());
			fprintf(header,"\t\t\treturn GL_FALSE;\n");
		}

		fprintf(header,"\t}\n");

		fprintf(header,"#endif //%s\n", extension_list[ext].name.c_str());
	}
	
	// else 
	fprintf(header, "\telse\n"
			"\t\treturn GL_FALSE;\n\n");

	fprintf(header,"\treturn GL_TRUE;\n");

	fprintf(header,"}\n\n");

	fprintf(header, "#endif %s_EXT_INIT_SOURCE\n\n", upfprefix.c_str());

//////////////////////////////////////////////////////////////////////////
// clean up macros and similar stuff
//////////////////////////////////////////////////////////////////////////
	// extern "c" end
	fprintf(header,"#ifdef __cplusplus\n");
	fprintf(header,"}\n");
	fprintf(header,"#endif\n\n");

	fprintf(header, "#undef EXTDECL\n" );

	fprintf(header,"#endif //__FILE_%s_EXT_PROC\n\n", upfprefix.c_str());
}

		
//////////////////////////////////////////////////////////////////////////
// "user" interface
//////////////////////////////////////////////////////////////////////////

void help()
{
printf("(W)GLextgen 1.2\n"
	   "by Andreas Kirsch 2003\n"
	   "\n"
	   "glextgen [-list] [-gl | -wgl] source output\n"
	   "\n"
	   "	-list			displays all extensions\n"
	   "	-gl or -wgl		type of the extensions\n"
	   "	source			source header\n"
	   "	output			output header file\n"
	   "\n"
	   "Default :\n"
	   " -gl  Source : .\\gl\\glext.h  Output : .\\gl\\glproc.h\n"
	   " -wgl Source : .\\gl\\wglext.h Output : .\\gl\\wglproc.h\n");

exit(0);
}
	   


int main(int argc, char *argv[])
{
#ifdef _DEBUG
//	_asm int 3
#endif
		
	string type;
	bool list = false;
	
	if(argc < 2 || argc > 5 || argv[1][0] != '-')
		help();

	ToLower(type = &argv[1][1]);
	
	if(type == "list")
	{
		list = true;
		
		if(argv[2][0] != '-')
			help();
		
		ToLower(type = &argv[2][1]);
	}
	
	if(type != "gl" && type != "wgl")
		help();

	Registry reg(type.c_str());
	
	if(argc > ((list) ? 3 : 2))
	{	
		reg.OpenHeader((list) ? argv[3] : argv[2]);

		reg.CreateOutputHeader((list) ? argv[4] :argv[3]);
	}
	else
	{
		if(type == "gl")
		{
			reg.OpenHeader(".\\gl\\glext.h");
			reg.CreateOutputHeader(".\\gl\\glproc.h");
		}
		else if(type == "wgl")
		{
			reg.OpenHeader(".\\gl\\wglext.h");

			reg.CreateOutputHeader(".\\gl\\wglproc.h");
		}
	}

	reg.Parse_Header();

	if(list)
		reg.List();

	reg.Write_Header();

	printf("\n+ Header written +\n");

	return 0;
}



	


	
	


